import math
import tributary.lazy as tl


class TestOps:
    def test_Negate(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        out = tl.Negate(t)
        assert out() == (-5, -1)

    def test_Invert(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        out = tl.Invert(t)
        assert out() == (1 / 5, -5**(-2))

    def test_Add(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        out = tl.Add(t, t)
        assert out() == (10, 2)

    def test_Sub(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        out = tl.Sub(t, t)
        assert out() == (0, 0)

    def test_Mult(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        out = tl.Mult(t, t)
        assert out() == (25, 10)

    def test_Div(self):
        t = tl.Node(value=(15, 1), use_dual=True)
        t2 = tl.Node(value=(5, 1), use_dual=True)
        out = tl.Div(t, t2)
        assert out() == (3, -10 / 25)

    def test_Pow(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        t2 = tl.Node(value=2, use_dual=True)
        out = tl.Pow(t, t2)
        assert out() == (25, 10)

    def test_Sum(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        out = tl.Sum(t, t, t, (3, 1))
        assert out() == (18, 4)

    def test_Average(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        out = tl.Average(t, (3, 1), (1, 1))
        assert out() == (3, 1)

    def test_Not(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        out = tl.Not(t)
        assert out() == False

    def test_And(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        out = tl.And(t, t)
        assert out() == (5, 1)

    def test_Or(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        out = tl.Or(t, t)
        assert out() == (5, 1)

    def test_Equal(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        out = tl.Equal(t, (5, 1))
        assert out()

    def test_NotEqual(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        out = tl.NotEqual(t, (7, 1))
        assert out()

    def test_Lt(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        out = tl.Lt(t, (10, 1))
        assert out()

    def test_Le(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        out = tl.Le(t, (5, 1))
        assert out()

    def test_Gt(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        out = tl.Gt(t, (5, 1))
        assert out() == False

    def test_Ge(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        out = tl.Ge(t, (5, 1))
        assert out()

    def test_Log(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        out = tl.Log(t)
        assert out() == (math.log(5), 1 / 5)

    def test_Sin(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        out = tl.Sin(t)
        assert out() == (math.sin(5), math.cos(5))

    def test_Cos(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        out = tl.Cos(t)
        assert out() == (math.cos(5), -math.sin(5))

    def test_Tan(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        out = tl.Tan(t)
        assert out() == (math.tan(5), (1 / math.cos(5))**2)

    def test_Arcsin(self):
        t = tl.Node(value=(0.5, 1), use_dual=True)
        out = tl.Arcsin(t)
        assert out() == (math.asin(0.5), 1 / math.sqrt(1 - 0.5**2))

    def test_Arccos(self):
        t = tl.Node(value=(0.5, 1), use_dual=True)
        out = tl.Arccos(t)
        assert out() == (math.acos(0.5), -1 / math.sqrt(1 - 0.5**2))

    def test_Arctan(self):
        t = tl.Node(value=(0.5, 1), use_dual=True)
        out = tl.Arctan(t)
        assert out() == (math.atan(0.5), 1 / (1 + 0.5**2))

    def test_Sqrt(self):
        t = tl.Node(value=(9, 1), use_dual=True)
        out = tl.Sqrt(t)
        assert out() == (3, 0.5 / math.sqrt(9))

    def test_Abs(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        out = tl.Abs(t)
        assert out() == (abs(5), 5 / abs(5))

    def test_Exp(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        out = tl.Exp(t)
        assert out() == (math.exp(5), math.exp(5))

    def test_Floor(self):
        t = tl.Node(value=(5.333, 1.333), use_dual=True)
        out = tl.Floor(t)
        assert out() == (math.floor(5.333), math.floor(1.333))

    def test_Ceil(self):
        t = tl.Node(value=(5.333, 1.333), use_dual=True)
        out = tl.Ceil(t)
        assert out() == (math.ceil(5.333), math.ceil(1.333))

    def test_Round(self):
        t = tl.Node(value=(5.333, 1.333), use_dual=True)
        out = tl.Round(t, ndigits=2)
        assert out() == (5.33, 1.33)

    def test_Int(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        out = tl.Int(t)
        assert out() == 5

    def test_Float(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        out = tl.Float(t)
        assert out() == 5.0

    def test_Bool(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        out = tl.Bool(t)
        assert out()

    def test_Str(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        out = tl.Str(t)
        assert out() == '5+1ε'

    def test_Len(self):
        t = tl.Node(value=(5, 1), use_dual=True)
        out = tl.Len(t)
        assert out() == 2



import aiohttp
import json as JSON

from .input import Foo


class WebSocket(Foo):
    '''Connect to websocket and yield back results

    Args:
        url (str): websocket url to connect to
        json (bool): load websocket data as json
        wrap (bool): wrap result in a list
    '''

    def __init__(self, url, json=False, wrap=False):
        async def _listen(url=url, json=json, wrap=wrap):
            session = aiohttp.ClientSession()
            async with session.ws_connect(url) as ws:

                async for msg in ws:
                    if msg.type == aiohttp.WSMsgType.TEXT:
                        x = msg.data
                        if json:
                            x = JSON.loads(x)
                        if wrap:
                            x = [x]
                        yield x
                    elif msg.type == aiohttp.WSMsgType.CLOSED:
                        break
                    elif msg.type == aiohttp.WSMsgType.ERROR:
                        break
        super().__init__(foo=_listen)
        self._name = 'WebSocket'



import asyncio
from .node import Node as StreamingNode  # noqa: F401
from .base import StreamingGraph  # noqa: F401
from .calculations import *  # noqa: F401, F403
from .control import *  # noqa: F401, F403
from .input import *  # noqa: F401, F403
from .output import *  # noqa: F401, F403
from .utils import *  # noqa: F401, F403
from ..base import StreamEnd


async def _run(node):
    out = Collect(node)  # noqa F405
    graph = out._construct_graph()
    nodes = graph.getNodes()

    value, last = None, None

    while True:
        for level in nodes:
            await asyncio.gather(*(asyncio.create_task(n()) for n in level))
        value, last = out.value(), value
        if isinstance(value, StreamEnd):
            break
    return last


def run(node):
    loop = asyncio.get_event_loop()
    if loop.is_running():
        # return future
        return asyncio.create_task(_run(node))
    # block until done
    return loop.run_until_complete(_run(node))



import tributary.streaming as ts
import pytest


class TestPostgres:
    @pytest.mark.skipif("int(os.environ.get('TRIBUTARY_SKIP_DOCKER_TESTS'))")
    def test_http(self):
        query = ['SELECT * FROM test']
        out = ts.PostgresSource(queries=query,
                                user='postgres',
                                database='postgres',
                                password='test',
                                host='localhost:5432')
        assert len(ts.run(out)) != 0



import math
import numpy as np
import scipy as sp
from .utils import _CALCULATIONS_GRAPHVIZSHAPE, _raise
from ..node import Node, _gen_node


def unary(foos, name):
    def _foo(self):
        foo = foos[0] if len(foos) == 1 or not self._use_dual else foos[1]
        downstream = Node(foo, {}, name=name, inputs=1, use_dual=self._use_dual, graphvizshape=_CALCULATIONS_GRAPHVIZSHAPE)
        self >> downstream
        return downstream
    return _foo


def binary(foos, name):
    def _foo(self, other):
        if not isinstance(other, Node):
            other = _gen_node(other)
            setattr(other, '_use_dual', self._use_dual)
        if self._use_dual != other._use_dual:
            raise NotImplementedError('Dual/Non-dual mismatch')
        foo = foos[0] if len(foos) == 1 or not self._use_dual else foos[1]
        downstream = Node(foo, {}, name=name, inputs=2, use_dual=self._use_dual, graphvizshape=_CALCULATIONS_GRAPHVIZSHAPE)
        self >> downstream
        other >> downstream
        return downstream
    return _foo


def n_ary(foos, name):
    def _foo(*others):
        use_dual = [x._use_dual for x in others]
        if np.any(use_dual) != np.all(use_dual):
            raise NotImplementedError('Dual/Non-dual mismatch')
        foo = foos[0] if len(foos) == 1 or not np.all(use_dual) else foos[1]
        downstream = Node(foo, {}, name=name, inputs=len(others), graphvizshape=_CALCULATIONS_GRAPHVIZSHAPE)
        for i, other in enumerate(others):
            other >> downstream
        return downstream
    return _foo


########################
# Arithmetic Operators #
########################
Noop = unary((lambda x: x,), name='Noop')
Negate = unary((lambda x: -1 * x, lambda x: (-1 * x[0], -1 * x[1])), name='Negate')
Invert = unary((lambda x: 1 / x, lambda x: (1 / x[0], -x[1] / (x[0]**2))), name='Invert')
Add = binary((lambda x, y: x + y, lambda x, y: (x[0] + y[0], x[1] + y[1])), name='Add')
Sub = binary((lambda x, y: x - y, lambda x, y: (x[0] - y[0], x[1] - y[1])), name='Sub')
Mult = binary((lambda x, y: x * y, lambda x, y: (x[0] * y[0], x[0] * y[1] + x[1] * y[0])), name='Mult')
Div = binary((lambda x, y: x / y, lambda x, y: (x[0] / y[0], (x[1] * y[0] - x[0] * y[1]) / y[0]**2)), name='Div')
RDiv = binary((lambda x, y: y / x, lambda x, y: (y[0] / x[0], (y[1] * x[0] - y[0] * x[1]) / x[0]**2)), name='RDiv')
Mod = binary((lambda x, y: x % y, lambda: _raise(NotImplementedError('Not Implemented!'))), name='Mod')
Pow = binary((lambda x, y: x ** y, lambda x, y: (x[0] ** y, y * x[1] * x[0] ** (y - 1))), name='Pow')
Sum = n_ary((lambda *args: sum(args), lambda *args: (sum([x[0] for x in args]), sum(x[1] for x in args))), name='Sum')
Average = n_ary((lambda *args: sum(args) / len(args), lambda *args: ((sum([x[0] for x in args]) / len(args), sum(x[1] for x in args) / len(args)))), name='Average')

#####################
# Logical Operators #
#####################
Not = unary((lambda x: not x,), name='Not')
And = binary((lambda x, y: x and y,), name='And')
Or = binary((lambda x, y: x or y,), name='Or')


###################
# Numpy Functions #
###################
def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
    if ufunc == np.add:
        if isinstance(inputs[0], Node):
            return inputs[0].__add__(inputs[1])
        else:
            return inputs[1].__add__(inputs[0])
    elif ufunc == np.subtract:
        if isinstance(inputs[0], Node):
            return inputs[0].__sub__(inputs[1])
        else:
            return inputs[1].__sub__(inputs[0])
    elif ufunc == np.multiply:
        if isinstance(inputs[0], Node):
            return inputs[0].__mul__(inputs[1])
        else:
            return inputs[1].__mul__(inputs[0])
    elif ufunc == np.divide:
        if isinstance(inputs[0], Node):
            return inputs[0].__truediv__(inputs[1])
        else:
            return inputs[1].__truediv__(inputs[0])
    elif ufunc == np.sin:
        return inputs[0].sin()
    elif ufunc == np.cos:
        return inputs[0].cos()
    elif ufunc == np.tan:
        return inputs[0].tan()
    elif ufunc == np.arcsin:
        return inputs[0].asin()
    elif ufunc == np.arccos:
        return inputs[0].acos()
    elif ufunc == np.arctan:
        return inputs[0].atan()
    elif ufunc == np.exp:
        return inputs[0].exp()
    elif ufunc == sp.special.erf:
        return inputs[0].erf()
    raise NotImplementedError('Not Implemented!')


def __array_function__(self, func, method, *inputs, **kwargs):
    if func == np.lib.scimath.log:
        return inputs[0][0].log()
    elif func == np.lib.scimath.sqrt:
        return inputs[0][0].sqrt()
    raise NotImplementedError('Not Implemented!')


###############
# Comparators #
###############
Equal = binary((lambda x, y: x == y, lambda x, y: x[0] == y[0]), name='Equal')
NotEqual = binary((lambda x, y: x != y, lambda x, y: x[0] != y[0]), name='NotEqual')
Lt = binary((lambda x, y: x < y, lambda x, y: x[0] < y[0]), name='Less')
Le = binary((lambda x, y: x <= y, lambda x, y: x[0] <= y[0] or x[0] == y[0]), name='LessOrEqual')
Gt = binary((lambda x, y: x > y, lambda x, y: x[0] > y[0]), name='Greater')
Ge = binary((lambda x, y: x >= y, lambda x, y: x[0] > y[0] or x[0] == y[0]), name='GreaterOrEqual')


##########################
# Mathematical Functions #
##########################
Log = unary((lambda x: math.log(x), lambda x: (math.log(x[0]), x[1] / x[0])), name='Log')
Sin = unary((lambda x: math.sin(x), lambda x: (math.sin(x[0]), math.cos(x[0]) * x[1])), name='Sin')
Cos = unary((lambda x: math.cos(x), lambda x: (math.cos(x[0]), -1 * math.sin(x[0]) * x[1])), name='Cos')
Tan = unary((lambda x: math.tan(x), lambda x: (math.tan(x[0]), x[1] * (1 / math.cos(x[0]))**2)), name='Tan')
Arcsin = unary((lambda x: math.asin(x), lambda x: (math.asin(x[0]), x[1] / math.sqrt(1 - x[0]**2))), name='Arcsin')
Arccos = unary((lambda x: math.acos(x), lambda x: (math.acos(x[0]), -1 * x[1] / math.sqrt(1 - x[0]**2))), name='Arccos')
Arctan = unary((lambda x: math.atan(x), lambda x: (math.atan(x[0]), x[1] / (1 + x[0]**2))), name='Arctan')
Sqrt = unary((lambda x: math.sqrt(x), lambda x: (math.sqrt(x[0]), x[1] * 0.5 / math.sqrt(x[0]))), name='Sqrt')
Abs = unary((lambda x: abs(x), lambda x: (abs(x[0]), x[1] * x[0] / abs(x[0]))), name='Abs')
Exp = unary((lambda x: math.exp(x), lambda x: (math.exp(x[0]), x[1] * math.exp(x[0]))), name='Exp')
Erf = unary((lambda x: math.erf(x), lambda x: _raise(NotImplementedError('Not Implemented!'))), name='Erf')
Floor = unary((lambda x: math.floor(x), lambda x: (math.floor(x[0]), math.floor(x[1]))), name='Floor')
Ceil = unary((lambda x: math.ceil(x), lambda x: (math.ceil(x[0]), math.ceil(x[1]))), name='Ceil')


def Round(self, ndigits=0):
    downstream = Node(lambda x: round(x, ndigits=ndigits) if not self._use_dual else
                      (round(x[0], ndigits=ndigits), round(x[1], ndigits=ndigits)),
                      {},
                      name="Round",
                      inputs=1,
                      graphvizshape=_CALCULATIONS_GRAPHVIZSHAPE)
    self.downstream().append((downstream, 0))
    downstream.upstream().append(self)
    return downstream


##############
# Converters #
##############
Int = unary((lambda x: int(x), lambda x: int(x[0])), name='Int')
Float = unary((lambda x: float(x), lambda x: float(x[0])), name='Float')
Bool = unary((lambda x: bool(x), lambda x: bool(x[0])), name='Bool')
Str = unary((lambda x: str(x), lambda x: str(x[0]) + '+' + str(x[1]) + 'ε'), name='Str')


# __Bool__ = unary(lambda x: bool(x), name='Noop')

def __Bool__(self):
    return True


###################
# Python Builtins #
###################
Len = unary((lambda x: len(x),), name='Len')


########################
# Arithmetic Operators #
########################
Node.__add__ = Add
Node.__radd__ = Add
Node.__sub__ = Sub
Node.__rsub__ = Sub
Node.__mul__ = Mult
Node.__rmul__ = Mult
# Node.__matmul__ = MatMult
# Node.__rmatmul__ = MatMult
Node.__div__ = Div
Node.__rdiv__ = RDiv
# Node.__divmod__ = DivMod
# Node.__rdivmod__ = DivMod
Node.__truediv__ = Div
Node.__rtruediv__ = RDiv
Node.__floordiv__ = Div
# Node.__lshift__ = LShift
# Node.__rlshift__ = LShift
# Node.__rshift__ = RShift
# Node.__rrshift__ = RShift

Node.__pow__ = Pow
Node.__rpow__ = Pow
Node.__mod__ = Mod
Node.__rmod__ = Mod

Node.sum = Sum
Node.average = Average
Node.invert = Invert


#####################
# Logical Operators #
#####################
Node.__and__ = And
Node.__rand__ = And
Node.__or__ = Or
Node.__ror__ = Or
# Node.__xor__ = Xor
# Node.__rxor__ = Xor
Node.__invert__ = Not
Node.__bool__ = __Bool__

##############
# Converters #
##############
Node.int = Int
# Node.__int__ = Int
Node.float = Float
# Node.__float__ = Float
# Node.__str__ = Str
# Node.__complex__ = Int

###############
# Comparators #
###############
Node.__lt__ = Lt
Node.__le__ = Le
Node.__gt__ = Gt
Node.__ge__ = Ge
Node.__eq__ = Equal
Node.__ne__ = NotEqual
Node.__neg__ = Negate
# Node.__pos__ =
# Node.__nonzero__ = Bool  # Py2 compat

###################
# Python Builtins #
###################
Node.__len__ = Len
# Node.__round__ = Len
# Node.__trunc__ = Len
# Node.__floor__ = Len
# Node.__ceil__ = Len

###################
# Numpy Functions #
###################
Node.__array_ufunc__ = __array_ufunc__
Node.__array_function__ = __array_function__


##########################
# Mathematical Functions #
##########################
Node.log = Log
Node.sin = Sin
Node.cos = Cos
Node.tan = Tan
Node.asin = Arcsin
Node.acos = Arccos
Node.atan = Arctan
Node.abs = Abs
Node.sqrt = Sqrt
Node.exp = Exp
Node.erf = Erf
Node.floor = Floor
Node.ceil = Ceil
Node.round = Round



import math
import numpy as np
import tributary.streaming as ts


rng = range(-10, 11)


def foo_range():
    for _ in rng:
        yield (_, 1)


pos_rng = range(1, 11)


def foo_pos():
    for _ in pos_rng:
        yield (_, 1)


neg_rng = range(-10, 0)


def foo_neg():
    for _ in neg_rng:
        yield (_, 1)


zero_one_rng = np.arange(0, 1, 0.05)  # [0,1)


def foo_zero_one():
    for _ in zero_one_rng:
        yield (_, 0.05)


class TestDualOps:
    def test_Noop(self):
        '''
        No-op
        '''
        t = ts.Timer(foo_range, count=len(rng), use_dual=True)
        out = ts.Noop(t)
        assert ts.run(out) == list(foo_range())

    def test_Negate(self):
        '''
        f = -x
        f' = -1
        '''
        expected_pos = [(-1 * x, -1) for x in pos_rng]
        expected_neg = [(-1 * x, -1) for x in neg_rng]
        t_pos = ts.Timer(foo_pos, count=len(pos_rng), use_dual=True)
        t_neg = ts.Timer(foo_neg, count=len(neg_rng), use_dual=True)
        out_pos = ts.Negate(t_pos)
        out_neg = ts.Negate(t_neg)
        assert ts.run(out_pos) == expected_pos
        assert ts.run(out_neg) == expected_neg

    def test_Invert(self):
        '''
        f = 1/x
        f' = -x^-2
        '''
        expected_pos = [(1 / x, -1 * x**(-2)) for x in pos_rng]
        expected_neg = [(1 / x, -1 * x**(-2)) for x in neg_rng]
        t_pos = ts.Timer(foo_pos, count=len(pos_rng), use_dual=True)
        t_neg = ts.Timer(foo_neg, count=len(neg_rng), use_dual=True)
        out_pos = ts.Invert(t_pos)
        out_neg = ts.Invert(t_neg)
        assert ts.run(out_pos) == expected_pos
        assert ts.run(out_neg) == expected_neg

    def test_Add(self):
        '''
        f = x+x
        f' = 2
        '''
        expected = [(x + x, 2) for x in rng]
        t = ts.Timer(foo_range, count=len(rng), use_dual=True)
        out = ts.Add(t, t)
        assert ts.run(out) == expected

    def test_Sub(self):
        '''
        f = x-x
        f' = 0
        '''
        expected = [(x - x, 0) for x in rng]
        t = ts.Timer(foo_range, count=len(rng), use_dual=True)
        out = ts.Sub(t, t)
        assert ts.run(out) == expected

    def test_Mult(self):
        '''
        f = x*x
        f' = 2x
        '''
        expected = [(x * x, 2 * x) for x in rng]
        t = ts.Timer(foo_range, count=len(rng), use_dual=True)
        out = ts.Mult(t, t)
        assert ts.run(out) == expected

    def test_Div(self):
        '''
        f = x/x
        f' = 0
        '''
        expected = [(1, 0) for x in pos_rng]
        t = ts.Timer(foo_pos, count=len(pos_rng), use_dual=True)
        out = ts.Div(t, t)
        assert ts.run(out) == expected

    def test_RDiv(self):
        '''
        f = x/x
        f' = 0
        '''
        expected = [(1, 0) for x in pos_rng]
        t = ts.Timer(foo_pos, count=len(pos_rng), use_dual=True)
        out = ts.RDiv(t, t)
        assert ts.run(out) == expected

    def test_Pow(self):
        '''
        f = x^2
        f' = 2x
        '''
        expected = [(x**2, 2 * x) for x in rng]
        t = ts.Timer(foo_range, count=len(rng), use_dual=True)
        out = ts.Pow(t, 2)
        assert ts.run(out) == expected

    def test_Sum(self):
        '''
        f = x+x+2
        f' = 2
        '''
        expected = [(x + x + 2, 2) for x in rng]
        t = ts.Timer(foo_range, count=len(rng), use_dual=True)
        t2 = ts.Timer(foo_range, count=len(rng), use_dual=True)
        c = ts.Const((2, 0), use_dual=True)
        out = ts.Sum(t, t2, c)
        assert ts.run(out) == expected

    def test_Average(self):
        '''
        f = (x + x + 1)/3
        f' = 2/3
        '''
        expected = [((x + x + 1) / 3, 2 / 3)for x in rng]
        t = ts.Timer(foo_range, count=len(rng), use_dual=True)
        t2 = ts.Timer(foo_range, count=len(rng), use_dual=True)
        c = ts.Const((1, 0), use_dual=True)
        out = ts.Average(t, t2, c)
        assert ts.run(out) == expected

    def test_Not(self):
        t = ts.Timer(foo_range, count=2, use_dual=True)
        out = ts.Not(t)
        assert ts.run(out) == [False, False]

    def test_And(self):
        t = ts.Timer(foo_range, count=2, use_dual=True)
        out = ts.And(t, t)
        assert ts.run(out) == [(-10, 1), (-9, 1)]

    def test_Or(self):
        t = ts.Timer(foo_range, count=2, use_dual=True)
        out = ts.Or(t, t)
        assert ts.run(out) == [(-10, 1), (-9, 1)]

    def test_Equal(self):
        t = ts.Timer(foo_range, count=2, use_dual=True)
        out = ts.Equal(t, t)
        assert ts.run(out) == [True, True]

    def test_NotEqual(self):
        t = ts.Timer(foo_range, count=2, use_dual=True)
        c = ts.Const((-10, 1), use_dual=True)
        out = ts.NotEqual(t, c)
        assert ts.run(out) == [False, True]

    def test_Lt(self):
        t = ts.Timer(foo_range, count=2, use_dual=True)
        c = ts.Const((1, 1), use_dual=True)
        out = ts.Lt(c, t)
        assert ts.run(out) == [False, False]

    def test_Le(self):
        t = ts.Timer(foo_range, count=2, use_dual=True)
        c = ts.Const((-9, 1), use_dual=True)
        out = ts.Le(c, t)
        assert ts.run(out) == [False, True]

    def test_Gt(self):
        t = ts.Timer(foo_range, count=2, use_dual=True)
        c = ts.Const((-9, 1), use_dual=True)
        out = ts.Gt(t, c)
        assert ts.run(out) == [False, False]

    def test_Ge(self):
        t = ts.Timer(foo_range, count=2, use_dual=True)
        c = ts.Const((-9, 1), use_dual=True)
        out = ts.Ge(t, c)
        assert ts.run(out) == [False, True]

    def test_Log(self):
        '''
        f = ln(x)
        f' = 1/x
        '''
        expected = [(math.log(x), 1 / x) for x in pos_rng]
        t = ts.Timer(foo_pos, count=len(pos_rng), use_dual=True)
        print(t)
        out = ts.Log(t)
        assert ts.run(out) == expected

    def test_Sin(self):
        '''
        f = sin(x)
        f' = cos(x)
        '''
        expected = [(math.sin(x), math.cos(x)) for x in rng]
        t = ts.Timer(foo_range, count=len(rng), use_dual=True)
        out = ts.Sin(t)
        assert ts.run(out) == expected

    def test_Cos(self):
        '''
        f = cos(x)
        f' = -sin(x)
        '''
        expected = [(math.cos(x), -1 * math.sin(x)) for x in rng]
        t = ts.Timer(foo_range, count=len(rng), use_dual=True)
        out = ts.Cos(t)
        assert ts.run(out) == expected

    def test_Tan(self):
        '''
        f = tan(x)
        f' = (1/cos(x))^2
        '''
        expected = [(math.tan(x), (1 / math.cos(x))**2) for x in rng]
        t = ts.Timer(foo_range, count=len(rng), use_dual=True)
        out = ts.Tan(t)
        assert ts.run(out) == expected

    def test_Arcsin(self):
        '''
        f = arcsin(x)
        f' = 1/sqrt(1-x^2)
        '''
        expected = [(math.asin(x), 0.05 / math.sqrt(1 - x**2)) for x in zero_one_rng]
        t = ts.Timer(foo_zero_one, count=len(zero_one_rng), use_dual=True)
        out = ts.Arcsin(t)
        assert ts.run(out) == expected

    def test_Arccos(self):
        '''
        f = arccos(x)
        f' = -1/sqrt(1-x^2)
        '''
        expected = [(math.acos(x), -0.05 / math.sqrt(1 - x**2)) for x in zero_one_rng]
        t = ts.Timer(foo_zero_one, count=len(zero_one_rng), use_dual=True)
        out = ts.Arccos(t)
        assert ts.run(out) == expected

    def test_Arctan(self):
        '''
        f = arctan(x)
        f' = 1/(1+x^2)
        '''
        expected = [(math.atan(x), 0.05 / (1 + x**2)) for x in zero_one_rng]
        t = ts.Timer(foo_zero_one, count=len(zero_one_rng), use_dual=True)
        out = ts.Arctan(t)
        assert ts.run(out) == expected

    def test_Sqrt(self):
        '''
        f = sqrt(x)
        f' = 0.5/sqrt(x)
        '''
        expected = [(math.sqrt(x), 0.5 / math.sqrt(x)) for x in pos_rng]
        t = ts.Timer(foo_pos, count=len(pos_rng), use_dual=True)
        out = ts.Sqrt(t)
        assert ts.run(out) == expected

    def test_Abs(self):
        '''
        f = abs(x)
        f' = x/abs(x)
        '''
        expected = [(abs(x), x / abs(x)) for x in neg_rng]
        t = ts.Timer(foo_neg, count=len(neg_rng), use_dual=True)
        out = ts.Abs(t)
        assert ts.run(out) == expected

    def test_Exp(self):
        '''
        f = exp(x)
        f' = exp(x)
        '''
        expected = [(math.exp(x), math.exp(x)) for x in rng]
        t = ts.Timer(foo_range, count=len(rng), use_dual=True)
        out = ts.Exp(t)
        assert ts.run(out) == expected

    def test_Floor(self):
        expected = [(math.floor(x), math.floor(0.05)) for x in zero_one_rng]
        t = ts.Timer(foo_zero_one, count=len(zero_one_rng), use_dual=True)
        out = ts.Floor(t)
        assert ts.run(out) == expected

    def test_Ceil(self):
        expected = [(math.ceil(x), math.ceil(0.05)) for x in zero_one_rng]
        t = ts.Timer(foo_zero_one, count=len(zero_one_rng), use_dual=True)
        out = ts.Ceil(t)
        assert ts.run(out) == expected

    def test_Round(self):
        expected = [(round(x, ndigits=1), round(0.05, ndigits=1)) for x in zero_one_rng]
        t = ts.Timer(foo_zero_one, count=len(zero_one_rng), use_dual=True)
        out = ts.Round(t, 1)
        assert ts.run(out) == expected

    def test_Int(self):
        t = ts.Timer(foo_range, count=len(rng), use_dual=True)
        out = ts.Int(t)
        assert ts.run(out) == list(rng)

    def test_Float(self):
        t = ts.Timer(foo_range, count=len(rng), use_dual=True)
        out = ts.Float(t)
        assert ts.run(out) == [float(x) for x in rng]

    def test_Bool(self):
        t = ts.Timer(foo_range, count=2, use_dual=True)
        out = ts.Bool(t)
        assert ts.run(out) == [True, True]

    def test_Str(self):
        t = ts.Timer(foo_range, count=1, use_dual=True)
        out = ts.Str(t)
        assert ts.run(out) == ['-10+1ε']

    def test_Len(self):
        t = ts.Timer(foo_range, count=1, use_dual=True)
        out = ts.Len(t)
        assert ts.run(out) == [2]



import os
import tributary.streaming as ts
import pytest


class TestFile:
    def test_file(self):
        file = os.path.abspath(os.path.join(os.path.dirname(__file__), 'test_file_data.csv'))
        if os.path.exists(file):
            os.remove(file)

        def foo():
            yield 1
            yield 2
            yield 3
            yield 4

        def read_file(file):
            with open(file, 'r') as fp:
                data = fp.read()
            return [int(x) for x in data]

        # Test that output is equal to what is read (generalized)
        out = ts.FileSink(ts.Foo(foo), filename=file)
        assert ts.run(out) == read_file(file)



import tributary.streaming as ts
import pytest


class TestHttp:
    @pytest.mark.skipif("int(os.environ.get('TRIBUTARY_SKIP_DOCKER_TESTS'))")
    def test_http(self):
        '''Test http server'''
        def foo():
            yield 'x'
            yield 'y'
            yield 'z'

        out = ts.HTTPSink(ts.Foo(foo), url='http://localhost:8080')
        assert len(ts.run(out)) == 3




class _Graph(object):
    '''internal representation of the entire graph state'''

    def __init__(self, node):
        self._starting_node = node

    def getNodes(self):
        return self._starting_node._deep_bfs()



import asyncio
import tributary.streaming as ts


class TestStreaming:
    def test_run_simple(self):
        t = ts.Const(value=1, count=1)
        assert ts.run(t) == [1]

    def test_run_foo(self):
        def foo():
            return 5
        t = ts.Foo(foo, count=1)
        assert ts.run(t) == [5]

    def test_run_generator(self):
        def foo():
            yield 1
            yield 2
        t = ts.Foo(foo)
        assert ts.run(t) == [1, 2]

    def test_run_async_foo(self):
        async def foo():
            await asyncio.sleep(.1)
            return 5

        t = ts.Foo(foo, count=1)
        assert ts.run(t) == [5]

    def test_run_async_generator(self):
        async def foo():
            yield 1
            yield 2
        t = ts.Foo(foo)
        assert ts.run(t) == [1, 2]

    def test_deep_bfs(self):
        a = ts.Const(1, count=1)
        b = ts.Random()
        c = ts.Curve([1, 2, 3])

        d = a + b
        e = a + c
        f = b + c

        g = ts.Print(d)
        h = ts.Print(e)
        i = ts.Print(f)

        def _ids(lst):
            return set([l._id for l in lst])

        def _ids_ids(lst_of_list):
            ret = []
            for lst in lst_of_list:
                ret.append(_ids(lst))
            return ret

        assert _ids(a._deep_bfs(tops_only=True)) == _ids([a, b, c])
        assert _ids(b._deep_bfs(tops_only=True)) == _ids([a, b, c])
        assert _ids(c._deep_bfs(tops_only=True)) == _ids([a, b, c])
        assert _ids(d._deep_bfs(tops_only=True)) == _ids([a, b, c])
        assert _ids(e._deep_bfs(tops_only=True)) == _ids([a, b, c])
        assert _ids(f._deep_bfs(tops_only=True)) == _ids([a, b, c])
        assert _ids(g._deep_bfs(tops_only=True)) == _ids([a, b, c])
        assert _ids(h._deep_bfs(tops_only=True)) == _ids([a, b, c])
        assert _ids(i._deep_bfs(tops_only=True)) == _ids([a, b, c])

        for x in (a, b, c, d, e, f, g, h, i):
            for y in (a, b, c, d, e, f, g, h, i):
                assert _ids_ids(x._deep_bfs()) == _ids_ids(y._deep_bfs())



import asyncpg
from ..node import Node


def Postgres(node, user, password, database, host, query_parser):
    '''Connects to Postgres and executes queries

        Args:
            user (str): postgres user
            password (str): postgres password
            database (str): postgres database
            host (str): postgres host
            query_parser (func): parse input node data to query list
    '''
    async def _send(data, query_parser=query_parser, user=user,
                    password=password, database=database, host=host):
        conn = await asyncpg.connect(user=user,
                                     password=password,
                                     database=database,
                                     host=host.split(':')[0],
                                     port=host.split(':')[1])
        queries = query_parser(data)
        for q in queries:
            await conn.execute(q)

        await conn.close()
        return data

    ret = Node(foo=_send, name='PostgresSink', inputs=1)
    node >> ret
    return ret



import asyncpg
import asyncio
from .input import Foo


class Postgres(Foo):
    '''Connects to Postgres and yields result of query

        Args:
            user (str): postgres user
            password (str): postgres password
            database (str): postgres database
            host (str): postgres host
            queries (str): list of queries to execute
            interval (int): seconds to wait before executing queries
            repeat (int): times to repeat
    '''

    def __init__(self, user, password, database, host,
                 queries, repeat=1, interval=1):
        async def _send(queries=queries, repeat=int(repeat),
                        interval=int(interval), user=user,
                        password=password, database=database, host=host):
            conn = await asyncpg.connect(user=user, password=password,
                                         database=database, host=host.split(':')[0],
                                         port=host.split(':')[1])
            count = 0 if repeat >= 0 else float('-inf')
            while count < repeat:
                data = []
                for query in queries:
                    values = await conn.fetch(query)
                    data.extend([list(x.items()) for x in values])
                yield data

                if interval:
                    await asyncio.sleep(interval)

                if repeat >= 0:
                    count += 1

            await conn.close()

        super().__init__(foo=_send)
        self._name = 'Postgres'





class StreamingGraph(object):
    def __init__(self, output_node):
        self._node = output_node

    def graph(self):
        return self._node.graph()

    def graphviz(self):
        return self._node.graphviz()

    def dagre(self):
        return self._node.dagre()

    def run(self):
        from tributary.streaming import run
        return run(self._node)



import tributary.streaming as ts
import pytest


class TestPostgres:
    @pytest.mark.skipif("int(os.environ.get('TRIBUTARY_SKIP_DOCKER_TESTS'))")
    def test_http(self):

        def foo():
            yield 1
            yield 2
            yield 3

        def parser(data):
            return ["INSERT INTO test(col1) VALUES ({});".format(data)]

        query = ['SELECT * FROM test']
        out = ts.PostgresSink(ts.Foo(foo),
                              query_parser=parser,
                              user='postgres',
                              database='postgres',
                              password='test',
                              host='localhost:5432')
        assert len(ts.run(out)) == 3



import tributary.streaming as ts
import pytest


class TestSocketIO:
    @pytest.mark.skipif("int(os.environ.get('TRIBUTARY_SKIP_DOCKER_TESTS'))")
    def test_socketio(self):
        '''Test socketio streaming'''
        def foo():
            yield 'a'
            yield 'b'
            yield 'c'

        out = ts.SocketIOSink(ts.Foo(foo), url='http://localhost:8069')
        assert ts.run(out) == ['a', 'b', 'c']




from .utils import _CONTROL_GRAPHVIZSHAPE
from ..node import Node


def If(if_node, satisfied_node, unsatisfied_node=None, *elseifs):
    '''Node to return satisfied if if_node else unsatisfied

    Args:
        if_node (Node): input booleans
        satisfied_node (Node): input to return if true
        unsatisfied_node (Node): input to return if False
        elseifs (Tuple(Node)): input of boolean/value pairs to evaluate as else if statements
    '''
    if len(elseifs) % 2 != 0:
        raise Exception('Else ifs must be in pairs')

    def foo(cond, if_, else_=None):
        # TODO else ifs
        if cond.value():
            return if_.value()
        return else_.value() if else_ is not None else None

    if isinstance(if_node._self_reference, Node):
        return if_node._gennode("If",
                                foo,
                                [if_node, satisfied_node, unsatisfied_node],
                                graphvizshare=_CONTROL_GRAPHVIZSHAPE)
    return if_node._gennode("If",
                            foo,
                            [if_node, satisfied_node, unsatisfied_node],
                            graphvizshare=_CONTROL_GRAPHVIZSHAPE)


Node.if_ = If



import aiohttp
import json as JSON
from .output import _OUTPUT_GRAPHVIZSHAPE
from ..node import Node
from ...base import StreamNone, StreamEnd


def WebSocket(node, url='', json=False, wrap=False, field=None, response=False, response_timeout=1):
    '''Connect to websocket and send data

    Args:
        foo (callable): input stream
        foo_kwargs (dict): kwargs for the input stream
        url (str): websocket url to connect to
        json (bool): dump data as json
        wrap (bool): wrap result in a list
    '''
    async def _send(data, url=url, json=json, wrap=wrap, field=field, response=response, response_timeout=response_timeout):
        if isinstance(data, (StreamNone, StreamEnd)):
            return data

        if wrap:
            data = [data]
        if json:
            data = JSON.dumps(data)

        session = aiohttp.ClientSession()
        async with session.ws_connect(url) as ws:
            await ws.send_str(data)

            if response:
                msg = await ws.receive(response_timeout)
                if msg.type == aiohttp.WSMsgType.TEXT:
                    x = msg.data
                elif msg.type == aiohttp.WSMsgType.CLOSED:
                    x = '{}'
                elif msg.type == aiohttp.WSMsgType.ERROR:
                    x = '{}'
            else:
                x = '{}'

        await session.close()

        if json:
            x = JSON.loads(x)

        if field:
            x = x[field]

        if wrap:
            x = [x]

        return x

    ret = Node(foo=_send, name='WebSocket', inputs=1, graphvizshape=_OUTPUT_GRAPHVIZSHAPE)
    node >> ret
    return ret
